#define TMAX(a,b) ((a)>(b)?(a):(b))

static	void		process_crontab(char *, char *,
					const char *, struct stat *,
					cron_db *, cron_db *);

//int debugmsg(int level,const char *format, ...);

void
load_database(cron_db *old_db) {
	struct stat statbuf, syscron_stat;
	cron_db new_db;
	DIR_T *dp;
	DIR *dir;
	user *u, *nu;

	Debug(DLOAD, ("[%ld] load_database()\n", (long)getpid()))

	/* before we start loading any data, do a stat on SPOOL_DIR
	 * so that if anything changes as of this moment (i.e., before we've
	 * cached any of the database), we'll see the changes next time.
	 */

	if (stat(spool_dir, &statbuf) < OK) {
		log_it("CRON", getpid(), "STAT FAILED", spool_dir);
		(void) exit(ERROR_EXIT);
	}

	/* track system crontab file
	 */
	if (stat(spool_dir, &syscron_stat) < OK)
		syscron_stat.st_mtime = 0;

	/* if spooldir's mtime has not changed, we don't need to fiddle with
	 * the database.
	 *
	 * Note that old_db->mtime is initialized to 0 in main(), and
	 * so is guaranteed to be different than the stat() mtime the first
	 * time this function is called.
	 */
	if (old_db->mtime == TMAX(statbuf.st_mtime, syscron_stat.st_mtime)) {
		Debug(DLOAD, ("[%ld] spool dir mtime unch, no load needed.\n",
			      (long)getpid()))
		return;
	}

	/* something's different.  make a new database, moving unchanged
	 * elements from the old database, reloading elements that have
	 * actually changed.  Whatever is left in the old database when
	 * we're done is chaff -- crontabs that disappeared.
	 */
	new_db.mtime = TMAX(statbuf.st_mtime, syscron_stat.st_mtime);
	new_db.head = new_db.tail = NULL;

//	if (syscron_stat.st_mtime)
//		process_crontab("root", NULL, SYSCRONTAB, &syscron_stat,
//				&new_db, old_db);

	/* we used to keep this dir open all the time, for the sake of
	 * efficiency.  however, we need to close it in every fork, and
	 * we fork a lot more often than the mtime of the dir changes.
	 */
	if (!(dir = opendir(spool_dir))) {
		log_it("CRON", getpid(), "OPENDIR FAILED", spool_dir);
		(void) exit(ERROR_EXIT);
	}

	while (NULL != (dp = readdir(dir))) {
		char fname[MAXNAMLEN+1], tabname[MAXNAMLEN+1];

		/* avoid file names beginning with ".".  this is good
		 * because we would otherwise waste two guaranteed calls
		 * to getpwnam() for . and .., and also because user names
		 * starting with a period are just too nasty to consider.
		 */
		if (dp->d_name[0] == '.')
			continue;

		if (strlen(dp->d_name) >= sizeof fname)
			continue;	/* XXX log? */
		(void) strcpy(fname, dp->d_name);
		
		if (!glue_strings(tabname, sizeof tabname, spool_dir,
				  fname, '/'))
			continue;	/* XXX log? */

		process_crontab(fname, fname, tabname,
				&statbuf, &new_db, old_db);
	}
	closedir(dir);

	/* if we don't do this, then when our children eventually call
	 * getpwnam() in do_command.c's child_process to verify MAILTO=,
	 * they will screw us up (and v-v).
	 */
	endpwent();

	/* whatever's left in the old database is now junk.
	 */
	Debug(DLOAD, ("unlinking old database:\n"))
	for (u = old_db->head;  u != NULL;  u = nu) {
		Debug(DLOAD, ("\t%s\n", u->name))
		nu = u->next;
		unlink_user(old_db, u);
		free_user(u);
	}

	/* overwrite the database control block with the new one.
	 */
	*old_db = new_db;
	Debug(DLOAD, ("load_database is done\n"))
}

void
link_user(cron_db *db, user *u) {
	if (db->head == NULL)
		db->head = u;
	if (db->tail)
		db->tail->next = u;
	u->prev = db->tail;
	u->next = NULL;
	db->tail = u;
}

void
unlink_user(cron_db *db, user *u) {
	if (u->prev == NULL)
		db->head = u->next;
	else
		u->prev->next = u->next;

	if (u->next == NULL)
		db->tail = u->prev;
	else
		u->next->prev = u->prev;
}

user *
find_user(cron_db *db, const char *name) {
	user *u;

	for (u = db->head;  u != NULL;  u = u->next)
		if (strcmp(u->name, name) == 0)
			break;
	return (u);
}

static void
process_crontab(char *uname, char *fname, const char *tabname,
		struct stat *statbuf, cron_db *new_db, cron_db *old_db)
{
	struct passwd *pw = NULL;
	int crontab_fd = OK - 1;
	user *u;
	int i,dot=0;
	char *tmp1;
	char *tmp2;
	char *fnametemp;
	int logjob=0;  //log or delete once task in fname.<log>.<jid>.log
	int idjob=0;   //jid for once task

	if (fname == NULL) {
		/* must be set to something for logging purposes.
		 */
		fname = "*system*";
	} else {
	//determine type of job: loop or once, log or delete.
	    dot=0;
	    for(i=0;i<strlen(fname);i++)
		if (fname[i]=='.') dot++;
	    if (dot==2) {  //once job
		fnametemp=malloc(strlen(uname));
		tmp1=malloc(strlen(uname));
		tmp2=malloc(strlen(uname));
		strcpy(fnametemp,uname);
		memset(uname,0,strlen(uname));
		memset(fname,0,strlen(fname));
		strcpy(fname,uname);
		sscanf(fnametemp,"%[^.].%[^.].%[^.]",uname,tmp1,tmp2);
		//printf(" %s %s %s\n",fname,tmp1,tmp2);
		if (!strcmp(tmp1,"log")) logjob=1;
		    idjob=atoi(tmp2);
		    free(tmp1);
		    free(tmp2);
		    free(fnametemp);
	    }
	    if ((pw = getpwnam(uname)) == NULL) {
			/* file doesn't have a user in passwd file.
			 */
			log_it(fname, getpid(), "ORPHAN", "no passwd entry");
			goto next_crontab;
		}
	}

	if ((crontab_fd = open(tabname, O_RDONLY|O_NONBLOCK|O_NOFOLLOW, 0)) < OK) {
		/* crontab not accessible?
		 */
		log_it(fname, getpid(), "CAN'T OPEN", tabname);
		goto next_crontab;
	}

	if (fstat(crontab_fd, statbuf) < OK) {
		log_it(fname, getpid(), "FSTAT FAILED", tabname);
		goto next_crontab;
	}
	if (!S_ISREG(statbuf->st_mode)) {
		log_it(fname, getpid(), "NOT REGULAR", tabname);
		goto next_crontab;
	}
	if ((statbuf->st_mode & 07777) != 0600) {
		log_it(fname, getpid(), "BAD FILE MODE", tabname);
		goto next_crontab;
	}
	if (statbuf->st_uid != ROOT_UID && (pw == NULL ||
	    statbuf->st_uid != pw->pw_uid || strcmp(uname, pw->pw_name) != 0)) {
		log_it(fname, getpid(), "WRONG FILE OWNER", tabname);
		goto next_crontab;
	}
	if (statbuf->st_nlink != 1) {
		log_it(fname, getpid(), "BAD LINK COUNT", tabname);
		goto next_crontab;
	}

	Debug(DLOAD, ("\t%s:", fname))
	u = find_user(old_db, fname);
	if (u != NULL) {
		/* if crontab has not changed since we last read it
		 * in, then we can just use our existing entry.
		 */
		if (u->mtime == statbuf->st_mtime) {
			Debug(DLOAD, (" [no change, using old data]"))
			unlink_user(old_db, u);
			link_user(new_db, u);
			goto next_crontab;
		}

		/* before we fall through to the code that will reload
		 * the user, let's deallocate and unlink the user in
		 * the old database.  This is more a point of memory
		 * efficiency than anything else, since all leftover
		 * users will be deleted from the old database when
		 * we finish with the crontab...
		 */
		Debug(DLOAD, (" [delete old data]"))
		unlink_user(old_db, u);
		free_user(u);
		log_it(fname, getpid(), "RELOAD", tabname);
	}
	u = load_user(crontab_fd, pw, fname);

	if (u != NULL) {
		u->mtime = statbuf->st_mtime;
		link_user(new_db, u);
	}

 next_crontab:
    if (dot==2) {
//	log_it(fname,getpid(),"SOMETHING NEED FOR REMOVE",tabname);
	if(logjob==1) unlink(tabname);
	else {  //rewrite me please
		char *tmp=malloc(MAXPATHLEN);
		memset(tmp,0,strlen(tmp));
		strcpy(tmp,tabname);
		strcat(tmp,".log");
		rename(tabname,tmp);
		free(tmp);
	}
    }

	if (crontab_fd >= OK) {
		Debug(DLOAD, (" [done]\n"))
		close(crontab_fd);
	}
}
